#!/usr/bin/perl
# Copyright (c) 2008, Rob Bloom
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

use strict;
use CGI;

my $query = new CGI;

my $chunklen = 0;		# Size of previous data to base next letter on
	# Comes from datafile, should not be set here.
my $softcap = 4;		# Start increasing value of end-word after here


my $maxcount = 0;		# Set in the script later
print <<EOF;
Content-type: text/html

<html>
<head><title>Random name/word generator</title></head>
<body>
<form action="">
<center>
<table width=400 border=1>
<tr><td colspan=2><center><h2>Random name/word generator</h2></center></td></tr>
<tr><td colspan=2>This is a script to generate random names or words based off a list of real names/words. The data comes from various publicly available sources (US census data, dictionary files, forum userlists, etc), and is fed into a script to read the probability of small chunks of text.  These probabilities are then used to randomly generate a statistically likely word based off that data.  Additionally, any randomly generated words are checked against their source data and discarded if it's found that the word is actually not random.</td></tr>
<tr><td>
	Select a dictionary to read from:<br>
	<select name="dict">
EOF
opendir(DIR, "dat/") || die "</select>Opps can't read data directory";
foreach (sort readdir DIR) {
	if (/^(.*)\.dat$/) {
		if ($1 eq $query->param('dict')) {
			print "    <option value=\"$1\" selected=\"selected\">$1</option>\n";
		} else {
			print "    <option value=\"$1\">$1</option>\n";
		}
	}
}
closedir DIR;
print <<EOF;
	</select>
</td><td>
	Select the number of words to generate:<br>
	<select name="count">
EOF
foreach ( (1,5,10,50,100,500,1000,5000,1e4) ) {
	my $x;
	$x = " selected=\"selected\"" if (($query->param('count') eq $_) || (!$query->param('count') && $_ == 100));
	print "    <option value=\"$_\"$x>$_</option>\n";
	$maxcount = $_ if ($_ > $maxcount);
}
print <<EOF;
	</select>
</td></tr>
<tr><td colspan="2">
<center><input type="submit" value="Randomize!"></center>
</td></tr>
</table>
</center>
<br><br>
EOF

my $datfile;
if ($datfile = $query->param('dict')) {
	open DAT, "dat/$datfile.dat" or die "Unable to open datafile!\n";
} else {
	print "</body></html>\n";
	exit 0;
}

my $count = $query->param('count');
$count = 1 if ($count < 1);
$count = $maxcount if ($count > $maxcount);

print "<b>$count random words from $datfile:</b><br>\n";
my %chunks;
my %realwords;

$_=<DAT>;
chomp;
if ($_ eq "__namegen namedata") {
	$chunklen = <DAT>;
	chomp $chunklen;
	print STDERR "Setting chunklen to $chunklen as per datafile's settings\n";
	READCHUNK: while (<DAT>) {
		chomp;
		last READCHUNK if ($_ eq "____words____");
		if (/^(.*)<:>(.*)<:>(.*)/) {
			$chunks{$1}{$2} = $3
		} else {
			print STDERR "Malformed line: $_\n";
		}
	}
	print STDERR "Chunks read, reading wordlist\n";
	while (<DAT>) {
		chomp;
		$realwords{$_} = 1;
	}
	print STDERR "Words read.\n";
} else {
	print "ERROR: Input source is not a datafile.  This version will not compute data on the fly!\n";
	exit(1);
}

my %donewords; # Words that we already used, to prevent dupes

while ($count) {
	my $chunk;
	for (my $i = 0; $i < $chunklen; $i++) {
		$chunk .= "_";
	}

	my $word = "";
	WORD: while (1) {
		my %next = %{ $chunks{$chunk} };
		if (length($word) > $softcap) {
			$next{0} *= length($word) - $softcap;
		}
		#print "Likely next characters for $chunk:\n";
		my $weight = 0;
		foreach (keys %next) {
			#print "    $_: " . $next{$_} . "\n";
			$weight += $next{$_};
		}
		my $r = int(rand($weight));
		#print "R: $r\n";
		CHAR: foreach (keys %next) {
			if ($r < $next{$_}) {
				if ($_ eq 0) {
					if ($donewords{$word} || length($word) <= $chunklen || $realwords{$word}) {
						#print "DUPE: $word\n";
						#sleep 1;
					} else {
						$donewords{$word} = 1;
						#print "$word\n";
						$count--;
					}
					last WORD;
				}
				$word .= $_;
				#print "Going with '$_', so far it is $word\n";
				$chunk = substr($chunk,1) . $_;
				last CHAR;
			} else {
				$r -= $next{$_};
				#print "R: $r\n";
			}
		}
	}
}


print "<ul>\n";
print "<li>$_<br>\n" foreach (sort (keys %donewords));
print "</ul>\n";
print "</body></html>\n";
exit 0;
