#!/usr/bin/perl
# Copyright (c) 2008, Rob Bloom
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

use strict;

my $writedat = 0;		# If true, write a file to quickly use the same data
my $chunklen = 2;		# Size of previous data to base next letter on
my $num = 10;			# Number of words to generate
my $softcap = 4;		# Start increasing value of end-word after here

print "Random name generator\n" if (!$writedat);
my %chunks;

my %realwords;

$_=<>;

chomp;
if ($_ eq "__namegen namedata") {
	$chunklen = <>;
	chomp $chunklen;
	print STDERR "Setting chunklen to $chunklen as per datafile's settings\n";
	READCHUNK: while (<>) {
		chomp;
		last READCHUNK if ($_ eq "____words____");
		if (/^(.*)<:>(.*)<:>(.*)/) {
			$chunks{$1}{$2} = $3
		} else {
			print STDERR "Malformed line: $_\n";
		}
	}
	print STDERR "Chunks read, reading wordlist\n";
	while (<>) {
		chomp;
		$realwords{$_} = 1;
	}
	print STDERR "Words read.\n";
} else {
	do {
		chomp;
		foreach (split /\W+/) {
			$realwords{lc $_} = 1;
			my $chunk;
			for (my $i = 0; $i < $chunklen; $i++) {
				$chunk .= "_";
			}

			foreach (split //) {
				$_=lc;
				$chunks{$chunk}{$_}++;
				$chunk = substr($chunk,1) . $_;
			}
			$chunks{$chunk}{0}++;
		}	
	} while (<>);
}

# foreach (keys %chunks) {
# 	my $c = $_;
# 	print "CHUNK $c\n";
# 	foreach (keys %{$chunks{$_}}) {
# 		print "    $_: " . $chunks{$c}{$_} . "\n";
# 	}
# }

if ($writedat) {
	print STDERR "Writing collected data to stdout.\n";
	print "__namegen namedata\n$chunklen\n";
	foreach (sort keys %chunks) {
		my $chnk = $_;
		foreach (sort keys %{$chunks{$chnk}}) {
			print "${chnk}<:>${_}<:>" . $chunks{$chnk}{$_} . "\n";
		}
	}
	print "____words____\n";
	foreach (keys %realwords) {
		print "$_\n";
	}
	exit(0);
}

print STDERR "Done reading data. Generating words, please hold...\n";

my %donewords;

print "Here are $num random names:\n";
while ($num) {
	my $chunk;
	for (my $i = 0; $i < $chunklen; $i++) {
		$chunk .= "_";
	}

	my $word = "";
	WORD: while (1) {
		my %next = %{ $chunks{$chunk} };
		if (length($word) > $softcap) {
			$next{0} *= length($word) - $softcap;
		}
		#print "Likely next characters for $chunk:\n";
		my $weight = 0;
		foreach (keys %next) {
			#print "    $_: " . $next{$_} . "\n";
			$weight += $next{$_};
		}
		my $r = int(rand($weight));
		#print "R: $r\n";
		CHAR: foreach (keys %next) {
			if ($r < $next{$_}) {
				if ($_ eq 0) {
					if ($donewords{$word} || length($word) <= $chunklen || $realwords{$word}) {
						#print "DUPE: $word\n";
						#sleep 1;
					} else {
						$donewords{$word} = 1;
						#print "$word\n";
						$num--;
					}
					last WORD;
				}
				$word .= $_;
				#print "Going with '$_', so far it is $word\n";
				$chunk = substr($chunk,1) . $_;
				last CHAR;
			} else {
				$r -= $next{$_};
				#print "R: $r\n";
			}
		}
	}
}


print "$_\n" foreach (sort (keys %donewords));
